
/* ================================================================== */
/*		NOTE GENERATION UNIT: Martin's Hopalong               */
/* ================================================================== */
/*
 * Hopalong attractor by Barry Martin - orbit in two dimensions.
 * 
 *       x(0) = y(0) = 0
 *       x(n+1) = y(n) - sign(x(n))*sqrt(abs(b*x(n)-c))
 *       y(n+1) = a - x(n)
 * 
 * Parameters are a, b, and c. The function "sign()"  returns 1 if the argument
 * is positive, -1 if argument is negative.
 * 
 * This fractal continues to develop in surprising ways after many iterations.
 * 
 * Another Martin fractal is simpler. The iterated formula is:
 * 
 *       x(n+1) = y(n) - sin(x(n))
 *       y(n+1) = a - x(n)
 * 
 * The parameter is "a". Try values near the number pi.
 */
/* ================================================================== */

typedef struct {
  FLOAT a,b,c;
  FLOAT x,y;
  FLOAT lx,hx;
  FLOAT sc,sq;
  int   ix,iy,isq,sg;
 } Mhop;

typedef struct {
  double a,b,c;
  double x,y;
  FLOAT  meanx,meany,maxx,maxy;
  FLOAT  sc,sq;
  int    ix,iy,aix,aiy;
  int    isq,sg;
  int    Rseed,Tofs;
  FLOAT  Fseed;
  unsigned Mvel;
 } Mhop2;

