
/* ================================================================== */
/*		NOTE GENERATION UNIT: Martin's Hopalong               */
/*			version 003                                   */
/* ================================================================== */

#include "../tsrc/mplay.h"
#include "../proteus.h"
#include "../moc1.h"
#include "martin.h"

#define KEYCENT		60
#define MINKEY		28
#define MAXKEY		90
#define MINVEL		21
#define MIDVEL		60
#define MAXVEL		110

#define MINDELTA	5
#define MAXDELTA	2000

#define LIMD_L		500
#define LIMD_H		5000

#define RTAB		((rand()&63)+24)

Mhop  Martpar;
int   N16, N32, Etb;

char Mode[128];

/* ================================================================== */

unsigned trill  (NoteEvent *N,unsigned t);
unsigned mtrill (NoteEvent *N,unsigned t);
unsigned scale  (NoteEvent *N,unsigned t);
unsigned pscale (NoteEvent *N,unsigned t);
unsigned mscale (NoteEvent *N,unsigned t);
unsigned fgroup (NoteEvent *N,unsigned t);
unsigned block  (NoteEvent *N,unsigned t);
unsigned melody (NoteEvent *N,unsigned t);
unsigned noplay (NoteEvent *N,unsigned t);

/* ================================================================== */

void martin(Mhop *M)
{
 FLOAT t,u;
 int f;

  t=M->b*M->x-M->c; if (t<0) t= -t;
  if (M->x<0) u=M->y+sqrt(t);
   else if (M->x>0) u=M->y-sqrt(t);
    else u=M->y;
  M->y = M->a-M->x;
  M->x = u;
  f=0;

  if (M->x<M->lx) { M->lx=M->x; f=1; }
   else if (M->x>M->hx) { M->hx=M->x; f=1; }
  if (f)
    M->sc = 100/(M->hx-M->lx);

  M->ix = M->x*M->sc;
  M->iy = M->y*M->sc;

  M->sq = sqrt(M->ix*M->ix+M->iy*M->iy);
}


unsigned inrange(int n, unsigned lb, unsigned hb)
{
  if (n<lb) {
    n += (((lb-n)/12+1)*12);
  }
  else if (n>hb) {
    n -= (((n-hb)/12+1)*12);
  }
  while (Mode[n]==0) n=(n+1)&127;
  return(n);
}


/* ================================================================== */


void InitGen(void)
{
 int i,j,k;
 float f;

  Martpar.a = HOSTP_F(0);
  Martpar.b = HOSTP_F(1);
  Martpar.c = HOSTP_F(2);
  Martpar.x = 0;
  Martpar.y = 0;
  Martpar.lx = 0;
  Martpar.hx = 0;
  Martpar.sc = HOSTP_I(0); if (Martpar.sc<=0) Martpar.sc=10;
  
  f = Martpar.a*Martpar.b*Martpar.c;
  f1init(f);
  _memcpy(&k,&f,4);
  Rseed=k;
  srand(k);

  if (f==0) f=500;
   else if (f<0) f = -f;
  while (f<500) f*=3;
  while (f>3000) f/=3;
  Btempo=f;
  N16=125000.0/Btempo;
  N32=62500.0/Btempo;

  if (HOSTP_I(1)) {
    for (i=0; i<16; i++) {
      k = HOSTP_I(1);
      for (j=7; j>=0; --j) {
        Mode[i*8+j] = ((k&15)!=0);
        k >>= 4;
      }
    }
    Mode[62] = 1;
  }
  else
    for (k=0; k<128; k++) Mode[k] = Params.Mtempl[k];

}


/* ================================================================== */


unsigned fgroup (NoteEvent *N,unsigned t)
{
 int i,nt,np,du,is,lastP,lng;
 unsigned k,t1;

  nt=4+(rand()%31);
  lng=rand()&1;
  np=N->pitch;

  du = N->duration/5;
  if (lng) {
    if (du<EIGHTH) du=EIGHTH; else if (du>QUARTER) du=QUARTER;
  }
  else {
    if (du<THIRTYSECOND) du=THIRTYSECOND; else if (du>SIXTEENTH) du=SIXTEENTH;
  }
  lastP=0;
  t1=t+10;

  is = (MINVEL+du)&127;
  if (is<MIDVEL) is=MIDVEL;
   else if (is>MAXVEL) is=MAXVEL;

  for (i=0; i<nt; i++) {
    N->pitch = (np+(int)onefrand(-6,6))&127;
    k = (N->pitch/2)&3;
    switch(k) {
      case 0: while (Mode[N->pitch]==0) N->pitch=(N->pitch+1)&127;
	      N->deltat = du;
	      N->duration = du+5;
	      break;
      case 1: while (Mode[N->pitch]==0) N->pitch=(N->pitch-1)&127;
	      N->deltat = du*2;
	      N->duration = (du*2)+5;
	      t1=t;
	      break;
      case 2: while (Mode[N->pitch]==0) N->pitch=(N->pitch+1)&127;
	      N->deltat = du/2;
	      N->duration = (du/2)+5;
	      break;
      case 3: while (Mode[N->pitch]==0) N->pitch=(N->pitch-1)&127;
	      N->deltat = du;
	      N->duration = du+5;
	      break;
     }
    while (N->pitch<MINRANGE(Etb)) N->pitch+=12;
    while (N->pitch>MAXRANGE(Etb)) N->pitch-=12;
    if (N->pitch!=lastP) t = note(N,t);
    lastP=N->pitch;
   }
  return(t1);
}


/* ================================================================== */


unsigned zcount=0;


void GenLoop(void)
{
 int Ntb,i,k,mxv,iter,Oldt,Tofs;
 unsigned t0,t1,maxt;		/* microsec */
 Mhop *M = &Martpar;
 NoteEvent N;

  mxv=0; N.pitch=0; N.a=0; N.b=0; N.c=0;
  N.startvol=N.deltavol=0;
  N.startpress=N.deltapress=0;
  N.lock=0;

  k=HOSTP_I(2); if (k<=4) k=4;
  for (i=0; i<k; i++) martin(M);

  asm { ldtimer; ldc 6; shl; ldc 10000; sum; st t0; }
  t0=ntempo(TimeD,t0);
  maxt=t0;
  iter=0; Oldt=0;
  Tofs = rand()&CHANMASK;

  for (;;) {

    martin(M);

    N.pitch = (M->ix+KEYCENT)&127;

    Ntb = ((Tofs&0x9)+(N.pitch&0x16)+(t0&0x30))&CHANMASK;

    if ((rand()%3)==0) {
      Ntb=Oldt; N.deltat=5;
     }
    else {
      N.deltat = ((Ntb&15)+1)*SIXTEENTH+(M->iy*32);
      if (N.deltat>MAXDELTA) N.deltat=MAXDELTA;
       else if (N.deltat<MINDELTA) N.deltat=MINDELTA;
     }
    Oldt=Ntb;

    k = ((int)(M->sq/4)%8)+(15-(Ntb&15));
    if ((Ntb<40)||(Ntb>47)) N.timbre=Ntb; else N.timbre=Ntb-24;
    N.duration = ((k%14)+1)*SIXTEENTH;

    N.intensity = ((25-k)*2)+MIDVEL;
    N.intensity += abs(N.pitch-KEYCENT)/2;
    if (N.intensity<MINVEL) N.intensity=MINVEL;
    if (N.intensity>MAXVEL) N.intensity=MAXVEL;
    if (N.intensity>mxv) mxv=N.intensity;

    N.pitch = inrange(N.pitch,MINRANGE(Ntb),MAXRANGE(Ntb));

    if (zcount==0) {
      t1 = t0+(N.duration*TimeD);
      if (t1>maxt) maxt=t1;		/* with duration */

      if (rand()&3)
	t0 = note(&N,t0);
      else
        fgroup(&N,t0);

      if (t0>maxt) maxt=t0;		/* with deltat */
     }

    iter++;

    if ((iter%11)==0) {
      k = RTAB%12;
      for (i=12; i<128; i+=12)
	Mode[i+k] ^= 1;
     }
    else if (HostCmd) {
      switch(HostCmd) {
        case 27:  return;
	case '>': zcount=100;
		  break;
	case 'R': for (k=0; k<100; k++) martin(M);
		  break;
	case 'F': TimeD /= 1.5; if (TimeD<LIMD_L) TimeD=LIMD_L;
		  t0=ntempo(TimeD,t0);
		  break;
	case 'D': TimeD *= 4; if (TimeD>LIMD_H) TimeD=LIMD_H;
		  t0=ntempo(TimeD,t0);
		  break;
	case 'G': TimeD /= 4; if (TimeD<LIMD_L) TimeD=LIMD_L;
		  t0=ntempo(TimeD,t0);
		  break;
	case 'S': TimeD *= 1.5; if (TimeD>LIMD_H) TimeD=LIMD_H;
		  t0=ntempo(TimeD,t0);
		  break;
	case 'Q': for (k=0; k<32; k++) t0=allnoteoff(k,t0+5);
		  break;
	case 'Z': M->x=0; M->y=0;
	case 'A': TimeD = Btempo;
		  t0=ntempo(TimeD,t0);
		  break;
	default:  M->x = HostCmd/128.0;
		  M->y = HostCmd/127.0;
		  Tofs = HostCmd&CHANMASK;
		  srand(Rseed);
		  break;
       }
      HostCmd=0;
     }
    if (zcount) zcount--;
     else
    asm { ld t0; ldc 6; shr; adc -100; tin; }

   }
}


/* ================================================================== */

