
/* ================================================================== */
/*		NOTE GENERATION UNIT: Martin's Hopalong               */
/*			version 005a                                  */
/* ================================================================== */

#include "../tsrc/mplay.h"
#include "../proteus.h"
#include "martin.h"

#define KEYCENT		64
#define MINKEY		28
#define MAXKEY		90
#define MINVEL		16
#define MIDVEL		48
#define MAXVEL		120
#define MINDELTA	5
#define MAXDELTA	2000

#define LIMD_L		150
#define LIMD_H		3000

#define RTAB		((rand()&63)+24)

Mhop  Martpar,MSH,MSH0;
int   N16;

int  Tbtab[NCHANS];

char Mode[128];


/* ================================================================== */
/*		FORMULA: Martin's Hopalong			      */
/* ================================================================== */

void martin(Mhop *M)
{
 float t,u;
 int f;

/* iterazione algoritmo */
  t=M->b*M->x-M->c; if (t<0) t= -t;
  if (M->x<0) u=M->y+sqrt(t);
   else if (M->x>0) u=M->y-sqrt(t);
    else u=M->y;
  M->y = M->a-M->x;
  M->x = u;

/* calcolo scala e correzione */
  f=0;
  if (M->x<M->lx) { M->lx=M->x; f=1; }
   else if (M->x>M->hx) { M->hx=M->x; f=1; }
  if (f)
    M->sc = 32/(M->hx-M->lx);

  M->ix = M->x*M->sc;
  M->sg = (M->ix==0? 0:(M->ix<0? -1:1));
  M->iy = M->y*M->sc;
  M->sg += (M->iy==0? 0:(M->iy<0? -1:1));

/* extra parameters */
  M->sq = sqrt(M->x*M->x+M->y*M->y);
  M->isq = sqrt(M->ix*M->ix+M->iy*M->iy);
}

/* ================================================================== */
/*		GENERATOR loop					      */
/*								      */
/*	Host parameter use:					      */
/*	F0 F1 F2		a,b,c				      */
/*	I0			scale factor			      */
/*	I1			mode rotation			      */
/*								      */
/* ================================================================== */

void InitGen(void)
{
 int i,k,r,t;
 float f;

  Martpar.a = HOSTP_F(0);
  Martpar.b = HOSTP_F(1);
  Martpar.c = HOSTP_F(2);
  Martpar.x = 0;
  Martpar.y = 0;
  Martpar.lx = 0;
  Martpar.hx = 0;
  Martpar.sc = HOSTP_I(0);

  f = Martpar.a*Martpar.b*Martpar.c;
  f1init(f);
  _memcpy(&k,&f,4);
  Rseed=k;
  srand(k);

  if (f==0) f=250;
   else if (f<0) f = -f;
  while (f<250) f*=3;
  while (f>3000) f/=3;
  Btempo=f;
  N16=125000.0/Btempo;

  for (k=0; k<NCHANS; k++) Tbtab[k]=k;
  for (k=0; k<NCHANS; k++) {
    r=rand()&CHANMASK;
    t=Tbtab[r]; Tbtab[r]=Tbtab[1]; Tbtab[1]=t;
   }
  MSH0=Martpar;
  MSH=Martpar;

  k=HOSTP_I(1);
  for (i=0; i<128; i++) {
    Mode[i] = Params.Mtempl[(i+k)&0x7f];
   }
}


unsigned inrange(int n, unsigned lb, unsigned hb)
{
  if (n<lb) {
    n += (((lb-n)/12+1)*12);
  }
  else if (n>hb) {
    n -= (((n-hb)/12+1)*12);
  }
  while (Mode[n]==0) n=(n+1)&127;
  return(n);
}


/* ================================================================== */


void GenLoop(void)
{
 int Ntb,i,k,iter,Tofs,Nofs;
 unsigned t0;
 Mhop *M = &Martpar;
 NoteEvent N,N1;

  N.startvol=N.deltavol=0;
  N.startpress=N.deltapress=0;
  N.pitch=0;

  for (i=0; i<4; i++) martin(M);
  Nofs=0;
  Tofs = rand()&CHANMASK;

  asm { ldtimer; ldc 6; shl; ldc 100000; sum; st t0; }
  t0=ntempo(TimeD,t0);
  iter=0;

  for (;;) {

    martin(M);

    oevent(M->ix,M->iy,t0);
    
    N.pitch = M->sq+Nofs+KEYCENT;

    Ntb = (Tofs+(M->ix%3)+M->sg)&CHANMASK;

    N.deltat = (Ntb%3)*EIGHTH;
    N.duration = (((M->iy%7)&7)+1)*SIXTEENTH;
    if (N.duration>(4*WHOLE)) N.duration=4*WHOLE;

    N.timbre=Ntb=Tbtab[Ntb];

    N.intensity = (unsigned)(M->isq%15)+MIDVEL;
    if (N.intensity<MINVEL) N.intensity=MINVEL;
     else if (N.intensity>MAXVEL) N.intensity=MAXVEL;

    N.pitch = inrange(N.pitch,MINRANGE(Ntb),MAXRANGE(Ntb));

    N.lock = NUNLOCK(Ntb);
    N1=N;
    if ((rand()%3) < 1) {
      N1.pitch = N.pitch+onefrand(-7,-3);
      fullnote(&N1,t0); t0+=5;
     }
    if ((rand()%3) < 2) {
      N1.pitch = N.pitch+onefrand(3,8);
      fullnote(&N1,t0); t0+=5;
     }
    t0 = fullnote(&N,t0);

    iter++;

    if ((iter%127)==0) {
      k = RTAB%9;
      for (i=12; i<128; i+=9)
	Mode[i+k] ^= 1;
     }
    else if (HostCmd) {
/* ================================================================== */
/*		host command handling				      */
/* ================================================================== */
      switch(HostCmd) {
	case 27:  return;
	case '.': Martpar=MSH;
		  srand(Rseed);
		  break;
	case ',': MSH=Martpar;
		  break;
	case '+': Nofs++; break;
	case '-': Nofs--; break;
	case '?': k = RTAB%12; if (k<2) k=12;
		  for (i=13; i<96; i+=k)
		    Mode[i] ^= 1;
		  break;
	case '/': for (i=13; i<96; i++)
		    Mode[i] ^= 1;
		  break;
	case 'R': for (k=0; k<100; k++) martin(M);
		  break;
	case 'G': TimeD /= 4; if (TimeD<LIMD_L) TimeD=LIMD_L;
		  t0=ntempo(TimeD,t0);
		  break;
	case 'F': TimeD /= 1.5; if (TimeD<LIMD_L) TimeD=LIMD_L;
		  t0=ntempo(TimeD,t0);
		  break;
	case 'Q': for (k=0; k<NCHANS; k++) t0=allnoteoff(k,t0+5);
		  break;
	case 'D': TimeD *= 4; if (TimeD>LIMD_H) TimeD=LIMD_H;
		  t0=ntempo(TimeD,t0);
		  break;
	case 'S': TimeD *= 1.5; if (TimeD>LIMD_H) TimeD=LIMD_H;
		  t0=ntempo(TimeD,t0);
		  break;
	case 'Z': Martpar=MSH0;
		  srand(Rseed);
	case 'A': TimeD = Btempo;
		  t0=ntempo(TimeD,t0);
		  break;
	default:  M->x = HostCmd/128.0;
		  M->y = HostCmd/127.0;
		  Tofs = HostCmd&CHANMASK;
		  srand(Rseed);
		  break;
/* ================================================================== */
       }
      HostCmd=0;
     }
    asm { ld t0; ldc 6; shr; adc -100; tin; }

   }
}


/* ================================================================== */

