
/* ================================================================== */
/*		NOTE GENERATION UNIT: Martin's Hopalong               */
/*			version 007                                   */
/* ================================================================== */

#include "../tsrc/mplay.h"
#include "../proteus.h"
#include "../canvas.h"
#include "../moc1.h"
#include "martin.h"

#define KEYCENT		60
#define MINKEY		28
#define MAXKEY		90
#define MINVEL		16
#define MIDVEL		32
#define MAXVEL		120

#define LIMD_L		300
#define LIMD_H		2000

#define RTAB		((rand()&63)+24)

#define R16		135000.0
#define R32		72500.0

Mhop  Martpar,MSH;
int   N16,N32,Pgrp;

int  Tbtab[NCHANS];

char Mode[128];

#define TRUENCHANS	NCHANS
#define TRUECHANMASK	CHANMASK

/* ================================================================== */


void martin(Mhop *M)
{
 float t,u;
 int f;

  t=M->b*M->x-M->c; if (t<0) t= -t;
  if (M->x<0) u=M->y+sqrt(t);
   else if (M->x>0) u=M->y-sqrt(t);
    else u=M->y;
  M->y = M->a-M->x;
  M->x = u;
  f=0;

  if (M->x<M->lx) { M->lx=M->x; f=1; }
   else if (M->x>M->hx) { M->hx=M->x; f=1; }
  if (f)
    M->sc = 32/(M->hx-M->lx);

  M->ix = M->x*M->sc;
  M->sg = (M->ix==0? 0:(M->ix<0? -1:1));
  M->iy = M->y*M->sc;
  M->sg += (M->iy==0? 0:(M->iy<0? -1:1));

  M->sq = sqrt(M->x*M->x+M->y*M->y);
  M->isq = sqrt(M->ix*M->ix+M->iy*M->iy);
}


/* ================================================================== */


void InitGen(void)
{
 int i,k,r,t;
 float f;

  Martpar.a = HOSTP_F(0);
  Martpar.b = HOSTP_F(1);
  Martpar.c = HOSTP_F(2);
  Martpar.x = 0;
  Martpar.y = 0;
  Martpar.lx = 0;
  Martpar.hx = 0;
  Martpar.sc = HOSTP_I(0);

  f = Martpar.a*Martpar.b*Martpar.c;
  f1init(f);
  _memcpy(&k,&f,4);
  Rseed=k;
  srand(k);

  if (f==0) f=LIMD_L;
   else if (f<0) f = -f;
  while (f<LIMD_L) f*=3;
  while (f>LIMD_H) f/=3;
  TimeD = Btempo =f;
  N16=R16/TimeD;
  N32=R32/TimeD;

  for (k=0; k<TRUENCHANS; k++) Tbtab[k]=k;
  for (k=0; k<TRUENCHANS; k++) {
    r=rand()&TRUECHANMASK;
    t=Tbtab[r]; Tbtab[r]=Tbtab[1]; Tbtab[1]=t;
   }

  k=HOSTP_I(1); if (k<1) k=1;
  for (i=0; i<128; i++) {
    Mode[i] = Params.Mtempl[(i*k)&0x7f];
   }
}


/* ================================================================== */


unsigned trill (int Ntb,NoteEvent *N,unsigned t)
{
 int i,nt,ht;

  if (Ntb&16) return(t);

  nt=(rand()%5)+1;
  ht=(rand()&1)+1;
  N->duration = N32-5;
  N->deltat = (N32+(rand()&31));

  for (i=0; i<nt; i++) {
    t = note(N,t);
    N->pitch += ht;
    t = note(N,t);
    N->pitch -= ht;
   }
  return(t);
}


unsigned scale (int Ntb,NoteEvent *N,unsigned t)
{
 int i,k,nt,sgn;

  if ((Ntb&16)==0) return(t);

  sgn = (rand()%3)-1; if (sgn==0) return(t);
  nt=1+(rand()%7);
  k=rand()&7;
  N->duration = ((N16*k)+(N32*(8-k)))/16;
  N->deltat = (N->duration+(rand()&31));

  for (i=0; i<nt; i++) {
    t = note(N,t);
    N->pitch += sgn*((rand()%3)+1);
   }
  return(t);
}


unsigned pscale (int Ntb,NoteEvent *N,unsigned t)
{
 int i,nt,np;

  if ((Ntb&16)==0) return(t);

  nt=12+(rand()%11);
  N->duration = N32-5;
  N->deltat = (N32+(rand()&15));
  np = (int)N->pitch-36;
  while (np<MINRANGE(Ntb)) np+=12;

  for (i=0; i<nt; i++) {
    np += onefrand(3,6);
    while (Mode[np]==0) np--;
    if (np>MAXRANGE(Ntb)) return(t);
    N->pitch=np;
    t = note(N,t);
   }
  return(t);
}


unsigned group (int Ntb,NoteEvent *N,unsigned t)
{
 int i,k,nt,np,lastP;

  nt=4+(rand()%26);
  np=N->pitch;

  if ((Ntb&16)==0) {
    k=rand()&7;
    N->duration = ((N16*k)+(N32*(8-k)))/16;
    N->deltat = (N->duration+(rand()&31));
   }
  else {
    N->duration = N32-5;
    N->deltat = (N32+(rand()&15));
   }
  lastP=0;

  for (i=0; i<nt; i++) {
    N->pitch = np+onefrand(-15,15);
    while (Mode[N->pitch]==0) N->pitch++;
    if (N->pitch!=lastP) t = note(N,t);
    lastP=N->pitch;
   }
  return(t);
}


unsigned block (NoteEvent *N,unsigned t)
{
 int i,k,np,is,mxd;
 unsigned t1;
 int durs[NCHANS];

  if (Pgrp==2) return(t);

  np=N->pitch;
  is=N->intensity;

  mxd=0;
  for (i=0; i<TRUENCHANS; i++) {
    durs[i]=((rand()%13)+1)*EIGHTH;
    if (durs[i]>mxd) mxd=durs[i];
   }
  N->deltat=(mxd/2)+10;

  k=rand()&0xd9ff;
  for (i=0; i<32; i++) {
    N->timbre=i;
    N->duration=durs[i];
    if (k&1) {
      if (rand()&1) {
	N->pitch=np+onefrand(-21,-3);
	while (Mode[N->pitch]==0) N->pitch++;
       }
      else {
	N->pitch=np+onefrand(3,21);
	while (Mode[N->pitch]==0) N->pitch--;
       }
      while (N->pitch<MINRANGE(i)) N->pitch+=12;
      while (N->pitch>MAXRANGE(i)) N->pitch-=12;
      N->intensity=is+onefrand(-20,5);
      t1=note(N,t); t+=5;
     }
    k>>=1;
   }
  k=rand()&0xff27;
  for (i=32; i<TRUENCHANS; i++) {
    N->timbre=i;
    N->duration=durs[i];
    if (k&1) {
      if (rand()&1) {
	N->pitch=np+onefrand(-21,-3);
	while (Mode[N->pitch]==0) N->pitch++;
       }
      else {
	N->pitch=np+onefrand(3,21);
	while (Mode[N->pitch]==0) N->pitch--;
       }
      while (N->pitch<MINRANGE(i)) N->pitch+=12;
      while (N->pitch>MAXRANGE(i)) N->pitch-=12;
      N->intensity=is+onefrand(-20,5);
      t1=note(N,t); t+=5;
     }
    k>>=1;
   }
  return(t1);
}


unsigned inrange(int n, unsigned lb, unsigned hb)
{
  if (n<lb) {
    n += (((lb-n)/12+1)*12);
  }
  else if (n>hb) {
    n -= (((n-hb)/12+1)*12);
  }
  while (Mode[n]==0) n=(n+1)&127;
  return(n);
}


/* ================================================================== */


void GenLoop(void)
{
 int Ntb,i,k,iter,Tofs,Nofs;
 int cond1,cond2,cond3;
 unsigned t0,t1;
 Mhop *M = &Martpar;
 NoteEvent N,N1;

  N.pitch=0; N.a=0; N.b=0; N.c=0;
  Pgrp=0; Nofs=0;
  N.startvol=N.deltavol=0;
  N.startpress=N.deltapress=0;
  N.pitch=0;
  N1=N;

  for (i=0; i<4; i++) martin(M);

  iter=0;
  Tofs = rand()&TRUECHANMASK;
  asm { ldtimer; ldc 6; shl; ldc 10000; sum; st t0; }
  t0=ntempo(TimeD,t0);

  for (;;) {

    martin(M);

    N.pitch = M->sq+Nofs+KEYCENT;

    Ntb = (Tofs+(M->iy%3)+M->sg)&TRUECHANMASK;
    if ((Ntb>39)&&(Ntb!=41)) Ntb=32+(Ntb&7);

    N.duration = EIGHTH+(Ntb<8? 0:((M->ix&3)*EIGHTH));
    N.deltat = N.duration+10;

    Ntb=Tbtab[Ntb];
    if ((Pgrp==1)&&(Ntb&16)) Ntb&=47;
     else if ((Pgrp==2)&&((Ntb&16)==0)) Ntb|=16;
    if (Ntb&16) Ntb &= 0x33;
    N.timbre=Ntb;

    N.intensity = (M->isq%31)+MIDVEL;
    if (N.intensity<MINVEL) N.intensity=MINVEL;
     else if (N.intensity>MAXVEL) N.intensity=MAXVEL;

    N.pitch = inrange(N.pitch,MINRANGE(Ntb),MAXRANGE(Ntb));

    N.lock = 0;
    N1=N;
    cond1=(M->ix&8);
    cond2=(M->iy&6);
    cond3=(M->isq&4);

    if (cond1) {
      N1.pitch = N.pitch+onefrand(-7,-3);
      while (Mode[N1.pitch]==0) N1.pitch--;
      N1.duration=N.duration/2;
      t1=note(&N1,t0);
      if (cond2||cond3) t0+=5; else t0=t1;
     }
    if (cond2) {
      N1.pitch = N.pitch+onefrand(3,8);
      while (Mode[N1.pitch]==0) N1.pitch++;
      N1.duration=N.duration*2;
      t1=note(&N1,t0);
      if (cond3) t0+=5; else t0=t1;
     }
    if (cond3) {
      if (N.pitch&1) {
	t0 = block(&N,t0);
       }
      else {
	k = rand()%11;
	switch(k) {
	  case 0:
	  case 1: t1 = trill(Ntb,&N,t0); break;
	  case 2: t1 = scale(Ntb,&N,t0); break;
	  case 3: t1 = group(Ntb,&N,t0); break;
	  case 4: t1 = pscale(Ntb,&N,t0); break;
	  default:t1 = note(&N,t0); break;
	 }
	if (t1==t0) t0+=(HALF*TimeD);
	 else t0=t1;
       }
     }

    iter++;

    if ((iter%17)==0) MSH=Martpar;

    if ((iter%43)==0) {
      Martpar=MSH;
      if (Pgrp!=2)
	for (k=0; k<16; k++) t0=allnoteoff(k,t0);
      k = RTAB%12;
      for (i=12; i<128; i+=12)
	Mode[i+k] ^= 1;
     }
    else if (HostCmd) {
      switch(HostCmd) {
        case 27:  return;
	case '+': Nofs++; break;
	case '-': Nofs--; break;
	case '?': k = (N.pitch+7)%12;
		  for (i=12; i<96; i+=12)
		    Mode[i+k] ^= 1;
		  break;
	case 'R': for (k=0; k<100; k++) martin(M);
		  break;
	case 'F': TimeD /= 1.5; if (TimeD<LIMD_L) TimeD=LIMD_L;
		  t0=ntempo(TimeD,t0);
		  break;
	case 'D': TimeD *= 4; if (TimeD>LIMD_H) TimeD=LIMD_H;
		  t0=ntempo(TimeD,t0);
		  break;
	case 'G': TimeD /= 4; if (TimeD<LIMD_L) TimeD=LIMD_L;
		  t0=ntempo(TimeD,t0);
		  break;
	case 'S': TimeD *= 1.5; if (TimeD>LIMD_H) TimeD=LIMD_H;
		  t0=ntempo(TimeD,t0);
		  break;
	case 'O': Pgrp=1; break;
	case 'P': Pgrp=2; break;
	case 'Q': for (k=0; k<16; k++) t0=allnoteoff(k,t0+5);
		  break;
	case 'Z': M->x=0; M->y=0;
		  M->lx=0; M->hx=0;
		  srand(Rseed);
		  _memcpy(Mode,Params.Mtempl,128);
		  Nofs=0;
	case 'A': TimeD = Btempo;
		  t0=ntempo(TimeD,t0);
		  Pgrp=0;
		  break;
	default:  M->x = HostCmd/1280.0;
		  M->y = HostCmd/1270.0;
		  Tofs = HostCmd&TRUECHANMASK;
		  srand(Rseed);
		  break;
       }
      HostCmd=0;
     }
    asm { ld t0; ldc 6; shr; adc -100; tin; }
   }
}


/* ================================================================== */

