
/* Included files */

#include <stdio.h>
#include <linux/link.h>

/* External procedures */

extern int Link;

/*
-- ----------------------------------------------------------------------------
--
--     Function Name : load_code
--
--     Input Parameters :
--	   char *boot_file_name - name of the boot file.
--
--     Output Parameters :
--	   None.
--
--     Result :
--	   int			- booting error code.
--
--     DESCRIPTION
--	   Boot the root transputer, ie. reset the root transputer, open the
--	   input file and send it to the tansputer as a stream of bytes.
--
-- ----------------------------------------------------------------------------
*/

static int boot_root (char *boot_file_name)
{
    FILE *boot_file;
    int result = 0;

    if ((boot_file = fopen(boot_file_name, "r")) != NULL)
    {
	int count = 0;
	char file_buffer[32000];

	while (result == 0)
	{
	    count = fread(file_buffer, 1, 32000, boot_file);
	    if (count > 0)
	    {
		if ( write( Link, &file_buffer, count ) != count )
			result = -2;
	    }
	    else if (feof(boot_file) == 0)
		result = -3;
	    else
		break;
	}
	if (result == 0)
	{
	    if (fclose(boot_file) != 0)
		result = -4;
	}
	else
	    (void) fclose(boot_file);
    }
    else
	result = -5;
    fprintf(stderr,"Booted Transputer, return %d\n", result);
    return(result);
}

/*
-- ----------------------------------------------------------------------------
--
--     Function Name : root_init
--
--     Input Parameters :
--	   char *boot_file_name	   - name of the boot file.
--	   int analyse_host	   - analyse transputer flag.
--	   int reset_host	   - reset transputer flag.
--
--     Output Parameters :
--	   long int *version_state - method used for transputer initialisation.
--	   char *core_memory	   - peeked memory from transputer (if any).
--
--     Result :
--	   int - initialisation error code.
--
--     DESCRIPTION
--	   Initialise the root transputer board by resetting it and sending the
--	   boot file to it.
--
-- ----------------------------------------------------------------------------
*/

int root_init (char *boot_file_name)
{
    if (!init_root())
    {
	if (boot_file_name != NULL)
	{
	    reset_root();

	    return (boot_root(boot_file_name));
	}
	else
	{
	    return(-2);
	}
    }
    return(-3);
}
