
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>

#define DEBUG
#undef  NO_OFF
#define NO_SYSEX
#undef  SMPTE
#define T8TIME
#define T8TDIV		4

#define TYPE_RAW	0
#define TYPE_ON		1
#define TYPE_OFF	2
#define TYPE_ALLOFF	3
#define TYPE_MARK	4

#ifdef DEBUG
#define Dprintf(s...)		printf(##s)
#else
#define Dprintf(s...)
#endif

/* -------------------------------------------------------------------- */

FILE *fs, *fd;
#ifdef NO_SYSEX
FILE *fx;
#endif
char Line[256];

typedef struct _event {
  struct _event *next;
  int t;
  unsigned char mpu,len;
  unsigned char data[32];
} EVENT;

EVENT *Tracks[64], *Ttail[64];
EVENT *ZeroTrk, *ZeroTtail;
int NEvTrk[64];
char *TrkNam[64];
unsigned char Cmpu[64],Cchan[64];
unsigned char scoreorder[64];

static int Tkeys[12] =
{
	0x56b,	/* 0101.01101011 -5 Dbmaj */
	0x5ab,	/* 0101.10101011 -4 Abmaj */
	0x5ad,	/* 0101.10101101 -3 Ebmaj */
	0x6ad,	/* 0110.10101101 -2 Bbmaj */
	0x6b5,	/* 0110.10110101 -1 Fmaj */
	0xab5,	/* 1010.10110101  0 Cmaj */
	0xad5,	/* 1010.11010101  1 Gmaj */
	0xad6,	/* 1010.11010110  2 Dmaj */
	0xb56,	/* 1011.01010110  3 Amaj */
	0xb5a,	/* 1011.01011010  4 Emaj */
	0xd5a,	/* 1101.01011010  5 Bmaj */
	0xd6a,	/* 1101.01101010  6 F#maj */
};

struct _htim {
  float t;
  int c;
} Histo[100];

float Hdiv[1024];
int Htone[12];

int Tonality;
int TimeUnit;
int OrigTimeMark=0;

/* -------------------------------------------------------------------- */

int hfc(const void *a, const void *b)
{
  struct _htim *p1 = (struct _htim *)a;
  struct _htim *p2 = (struct _htim *)b;
  return (p2->c - p1->c);
}

int getnote(char *s, unsigned char *pitch, unsigned char *vol)
{
  static int ix[7] = { 9,11,0,2,4,5,7 };
  int v, vo;

  v=ix[toupper(*s)-'A'];
  if (s[1]=='#') v++; else if (s[1]=='b') v--;
  if (s[2]=='-') return 0;
  v = (v%12) + (s[2]-'0'+2)*12;
  if (pitch) *pitch = v;
  if (vol) {
  	if (sscanf(s+3,"%d",&vo)==1) {
  	  if ((vo<0)||(vo>127)) vo&=0x7f;
  	}
  	else vo = 0;
  	*vol = vo;
  }
  return v;
}


EVENT *addevt(int trk, int t, int type, char *s)
{
  static int lastch=-1, lasttb=-1;
  EVENT *E = calloc(1,sizeof(EVENT));
  EVENT *E2;
  int i;

  if (E==NULL) return NULL;
  E->t = t;
  E->mpu = Cmpu[trk];
  switch(type) {
    case TYPE_ON:	E->len=3;
    			E->data[0]=0x90|(Cchan[trk]&15);
    			if (getnote(s,&E->data[1],&E->data[2]))
			  NEvTrk[trk]++;
			else {
    			  free(E);
    			  return NULL;
			}
    			break;
    case TYPE_OFF:	E->len=3;
#ifdef NO_OFF
    			E->data[0]=0x90|(Cchan[trk]&15);
#else
    			E->data[0]=0x80|(Cchan[trk]&15);
#endif
    			if (!getnote(s,&E->data[1],&E->data[2])) {
    			  free(E);
    			  return NULL;
			}
    			break;
    case TYPE_ALLOFF:	E->len=3;
    			E->data[0]=0xb0|(Cchan[trk]&15);
    			E->data[1]=0x7b;
    			E->data[2]=0;
    			break;
    case TYPE_RAW:	i=0;
    			while (*s && (i<32)) {
    			  int j;
    			  if (sscanf(s,"%x",&E->data[i])!=1) break;
    			  for (j=0; j<3; j++) { if (*s) s++; }
    			  i++;
    			}
    			E->len=i;
    			if ((E->data[0]==0xff)&&(E->data[1]==0x51)) {
    			  int k;
    			  k = (E->data[3]<<16)|(E->data[4]<<8)|(E->data[5]);
    			  if (OrigTimeMark==0) {
    			  	/* T8time = Btempo*500 */
    			  	OrigTimeMark = k/T8TDIV;
    			  }
 			  Dprintf("T8 quarter tempo = %d us\n",k/4);
    			  free(E);
    			  return NULL;
    			}
#ifdef NO_SYSEX
			if (E->data[0]==0xf0) {
			  if (fx) {
			    fputc(0xff,fx); fputc(0x21,fx); fputc(1,fx);
			    fputc(E->mpu,fx);
			    fputc(0xf0,fx);
			    fwrite(E->data+2,1,E->len-2,fx);
			  }
    			  free(E);
    			  return NULL;
			}
#endif
			if ((E->data[0]&0xf0)==0xc0) {
			  Dprintf("PGM_CHANGE ch%d = %d\n",(E->data[0]&15),
			  	E->data[1]);
			  E2 = calloc(1,sizeof(EVENT));
			  if (E2==NULL) return NULL;
			  E2->t = t;
			  E2->mpu = Cmpu[trk];
			  E2->data[0]=0xff; E2->data[1]=0x01; E2->data[2]=7;
			  sprintf(E2->data+3,"PGM %03d",E->data[1]);
			  E2->len=10;
			  Ttail[trk]->next = E2;
			  Ttail[trk] = E2;
			}
    			if ((E->len==8) && ((E->data[0]&0xf0)==0xb0) &&
    			    ((E->data[3]&0xf0)==0xb0) &&
    			    ((E->data[6]&0xf0)==0xc0)) {
			  if (((E->data[0]&15)!=lastch) ||
			      (E->data[7]!=lasttb)) {
    			  	E->len=10;
    			  	lasttb=E->data[9]=E->data[7];
    			  	E->data[8]=E->data[6];
    			  	E->data[7]=0;
    			  	E->data[6]=E->data[5];
    			  	E->data[5]=E->data[4];
    			  	E->data[4]=E->data[3];
    			  	E->data[3]=0;
    			  	lastch=(E->data[0]&15);
			 	Dprintf("PGM_CHANGE ch%d = 00:%d 32:%d %d\n",
			 		(E->data[0]&15),E->data[2],E->data[6],
			  		E->data[9]);
				E2 = calloc(1,sizeof(EVENT));
				if (E2==NULL) return NULL;
				E2->t = t;
				E2->mpu = Cmpu[trk];
				E2->data[0]=0xff; E2->data[1]=0x01; E2->data[2]=13;
				sprintf(E2->data+3,"PGM %02d %02d %03d",
					E->data[2],E->data[6],E->data[9]);
				E2->len=16;
				Ttail[trk]->next = E2;
				Ttail[trk] = E2;
    			  }
    			  else {
    			  	free(E);
    			  	return NULL;
    			  }
    			}
    			break;
  }
  Ttail[trk]->next = E;
  Ttail[trk] = E;
  return E;
}


/* -------------------------------------------------------------------- */


int pass1(int trkno)
{
  int lastoff;
  char *p;
  float t0,t,dt,ot0,ot,odt;

  t0 = ot0 = -1.0;
  odt = 0;
  lastoff = 0;

  while (fgets(Line,250,fs)) {
    if (*Line=='#') continue;
    if (strlen(Line)<2) return 0;
    p = strtok(Line,"\t\r\n ");
    if ((p==NULL) || (sscanf(p,"%f",&t)<1)) return 1;
    strtok(NULL,"\t\r\n ");
    p = strtok(NULL,"\r\n");
    while (isspace(*p)) p++;

    if (!strncmp("ON ",p,3)) {
        if (t0 < 0) t0 = t; if (t < t0) t = t0;
        dt = (t-t0)*1000.0; t0 = t;
        ot = t;
        if (ot0 < 0) ot0 = ot;
        if (ot < ot0) ot = ot0;
        odt = (ot-ot0)*100.0;		/* 10 msec */
        if ((odt>=4.0)&&(odt<40.0)) {
          int k = (int)odt;
          Histo[k].t = odt;
          Histo[k].c += 1;
        }
        ot0 = ot;
        Htone[getnote(p+4,NULL,NULL)%12] += 1;
        addevt(trkno-1,t*1000.0,TYPE_ON,p+4);
        lastoff=0;
    }
    else if (!strncmp("OFF ",p,4)) {
        if (t0 < 0) t0 = t; if (t < t0) t = t0;
        dt = (t-t0)*1000.0; t0 = t;
        addevt(trkno-1,t*1000.0,TYPE_OFF,p+4);
        lastoff=0;
    }
    else if (!strncmp("OFF>",p,4) && (lastoff==0)) {
        if (t0 < 0) t0 = t; if (t < t0) t = t0;
        dt = (t-t0)*1000.0; t0 = t;
        addevt(trkno-1,t*1000.0,TYPE_ALLOFF,p+4);
        lastoff++;
    }
    else if (isxdigit(p[0]) && isxdigit(p[1])) {
        if (t0 < 0) t0 = t; if (t < t0) t = t0;
        dt = (t-t0)*1000.0; t0 = t;
        addevt(trkno-1,t*1000.0,TYPE_RAW,p);
        lastoff=0;
    }
    Dprintf("dtime=%10.3f\t%10.3f\t",dt,odt*10.0);
    if (p) Dprintf("[%s]",p); Dprintf("\n");
  }
  return 1;
}


int getzerotrk(FILE *fs)
{
  char *p,*q;
  EVENT *E;
  int i;
  float t0;

  while (fgets(Line,250,fs)) {
    if (*Line=='#') continue;
    if (strlen(Line)<2) return 0;
    if ((p=strstr(Line,"MARKER"))) {
      if (!isalpha(p[7])) continue;
      if (sscanf(Line,"%f",&t0)!=1) return 1;
      E = calloc(1,sizeof(EVENT));
      E->t = t0*1000.0;
      E->data[0]=0xff; E->data[1]=0x01;
      p+=7; q=E->data+3; i=0;
      while (!isspace(*p) && (i<30)) *q++ = *p++, i++;
      while (isspace(*p)) p++;
      *q++ = ' ', i++;
      while (!isspace(*p) && (i<30)) *q++ = *p++, i++; *q=0;
      E->data[2] = i;
      E->len = i+3;
      Dprintf("Marker evt=%s\n",E->data+3);
      ZeroTtail->next = E;
      ZeroTtail = E;
    }
    else if ((p=strstr(Line,"COMMENT"))) {
      if ((p[8] != '"') || (p[11] == '=')) continue;
      if (sscanf(Line,"%f",&t0)!=1) return 1;
      E = calloc(1,sizeof(EVENT));
      E->t = t0*1000.0;
      E->data[0]=0xff; E->data[1]=0x01; E->data[2]=1;
      E->data[3]=p[9]&0x7f;
      E->len = i+3;
      Dprintf("Marker key=%c\n",E->data[3]);
      ZeroTtail->next = E;
      ZeroTtail = E;
    }
  }
  return 0;
}


/* -------------------------------------------------------------------- */


static void mwnum(int n, int siz)
{
  char *p = (char *)&n;
  while (siz--) {
    fputc(p[siz],fd);
  }
}

static unsigned int cksum = 0;

static int wdeltat(unsigned int t1,unsigned int t0,double dt)
{
  static double rest=0;
  double d, id;
  unsigned char b,c[5];
  unsigned int n;
  int i,g;

  if (dt==0.0) { rest=0.0; return 0; }

  d = ((double)(t1-t0) / dt) + rest;
  rest = modf(d, &id);
  if (rest >= 0.6) { id += 1.0; rest -= 1.0; }
  n = (int)id;
  cksum += n;

  Dprintf("T=%08d CK=%08d DT=%08d R=%8.5f",t1,(int)(cksum*dt+0.5),n,rest);
  for (i=4; i>=0; --i) {
    c[i]=n&0x7f; n>>=7;
  }
  g=0;
  for (i=0; i<5; i++) {
    if (c[i] || g || (i==4)) {
      if (i!=4) b=c[i]|0x80; else b=c[i];
      Dprintf(" %02x",b);
      fputc(b,fd);
      g++;
    }
  }
  return g;
}


int genmidi(void)
{
  unsigned char uN[128];
  EVENT *p, *Z;
  double Rtick;
  int i,j,k,t0,nwr,trkc,nsco,ok;
  unsigned long fptr;
  unsigned char Rstatus;

  fwrite("MThd",1,4,fd);
  mwnum(6,4);
  mwnum(1,2);
  k=0;
  for (nsco=0; nsco<64; nsco++) {
    i = scoreorder[nsco];
    if ((i<64) && (NEvTrk[i]!=0)) k++;
  }
  Dprintf("MThd: %d tracks\n",k);
  mwnum(k,2);

#ifdef SMPTE
  /* resolution 0.5 ms */
  fputc(-25,fd);
  fputc(80,fd);
  Rtick = 0.5;
  Dprintf("Resol 1 tick = 0.5 ms SMPTE\n");
#else
  /* resolution in ticks per quarter, 1 quarter = 8*TimeUnit ms */
  mwnum(96,2);
#ifdef T8TIME
  if (OrigTimeMark)
    Rtick = OrigTimeMark/96000.0;
  else
#endif
    Rtick = (8.0*TimeUnit)/96;
#endif
  Dprintf("Resol 1 tick = %g ms\n",Rtick);

  trkc = 0;
  for (nsco=0; nsco<64; nsco++) {
    i = scoreorder[nsco];
    if ((i>=64) || (NEvTrk[i]==0)) continue;
    p = Tracks[i]->next;
    Z = ZeroTrk->next;
    Dprintf("\nMTrk %d\n",trkc);
    fwrite("MTrk",1,4,fd);
    fptr=ftell(fd);
    mwnum(0,4);
    t0 = -1;
    nwr = 0;

    fputc(0,fd);
    fputc(0xff,fd); fputc(0x21,fd); fputc(1,fd);
    fputc(Cmpu[i],fd); nwr+=5;
    Dprintf("MPU %d\n",Cmpu[i]);
    fputc(0,fd);
    fputc(0xff,fd); fputc(0x20,fd); fputc(1,fd);
    fputc(Cchan[i],fd); nwr+=5;
    Dprintf("Chan %d\n",Cchan[i]);
    if (trkc==0) {
      fputc(0,fd);
      fputc(0xff,fd); fputc(0x59,fd); fputc(2,fd);
      fputc(Tonality&0xff,fd);
      fputc(0,fd); nwr+=6;
      Dprintf("Tonality %d\n",Tonality);
      fputc(0,fd);
      fputc(0xff,fd); fputc(0x51,fd); fputc(3,fd);
#ifdef T8TIME
      if (OrigTimeMark)
        k=OrigTimeMark;
      else
#endif
        k=8000*TimeUnit;
      fputc((k>>16)&0xff,fd);
      fputc((k>>8)&0xff,fd);
      fputc(k&0xff,fd); nwr+=7;
      Dprintf("Quarter %d us\n",k);
      fputc(0,fd);
      fputc(0xff,fd); fputc(0x58,fd); fputc(4,fd);
      fputc(4,fd); fputc(2,fd);		/* 4/4 */
      fputc(24,fd); fputc(8,fd); nwr+=8;
      Dprintf("Tempo 4/4,24,8\n");
    }
    if (TrkNam[i]) {
      int l = strlen(TrkNam[i]);
      if (l>64) l=64;
      if (l>0) {
        fputc(0,fd);
        fputc(0xff,fd); fputc(3,fd); fputc(l,fd);
        fwrite(TrkNam[i],1,l,fd);
        nwr += (l+4);
        fputc(0,fd);
        fputc(0xff,fd); fputc(4,fd); fputc(l,fd);
        fwrite(TrkNam[i],1,l,fd);
        nwr += (l+4);
        Dprintf("Track name [%s]\n",TrkNam[i]);
      }
    }

    memset(uN,0,128);
    Rstatus = 0;
    wdeltat(0,0,0.0);

    while (p) {
      int pd1=p->data[1];
      if (t0 < 0) { t0 = p->t; cksum = t0/Rtick; }
      ok=1;

      while (Z && Z->next && (Z->next->t < t0)) Z=Z->next;
      if (Z && (Z->t < t0)) {
	Dprintf("T=%08d DT=xxxxxxxx 00",Z->t);
	fputc(0,fd); nwr++;
        for (k=0; k < Z->len; k++) {
          Dprintf(" %02x",Z->data[k]);
          fputc(Z->data[k],fd);
          nwr++;
        }
	Dprintf("\n");
        Z = Z->next;
      }

      switch (p->data[0]&0xf0) {
        case 0x80:	/* note off */
#ifdef NO_OFF
		p->data[0]|=0x10; p->data[2]=0;
#endif
        	if (uN[pd1]==0) ok=0;
        	  else uN[pd1] = 0;
        	break;
        case 0x90:	/* note on */
#ifdef NO_OFF
		if (p->data[2]==0) {
			if (uN[pd1]==0) ok=0; else uN[pd1] = 0;
		}
		else {
			if (uN[pd1]) ok=0; else uN[pd1] = 1;
		}
#else
        	if (uN[pd1]) ok=0;
        	  else uN[pd1] = 1;
#endif
        	break;
        case 0xb0:
        	if (pd1==0x7b) {
        	  int u=0;
        	  for (j=0; j<128; j++) {
        	    if (uN[j]) {
        	      if (u) {
			Dprintf("T=%08d DT=%08d 00",p->t,0);
        	        fputc(0,fd); nwr++;
        	      }
        	      else
		        { nwr += wdeltat(p->t,t0,Rtick); u=1; }
#ifdef NO_OFF
        	      k = 0x90|(p->data[0]&15); 
        	      if (k!=Rstatus) {
		        Dprintf(" %02x",k);
        	      	fputc(k,fd); nwr++;
        	      }
        	      else {
		        Dprintf(" *");
        	      }
        	      Rstatus=k;
		      Dprintf(" %02x 00\n",j);
#else
        	      fputc(0x80|(p->data[0]&15),fd); nwr++;
		      Dprintf(" %02x %02x 00\n",0x80|(p->data[0]&15),j);
#endif
        	      fputc(j,fd); fputc(0,fd); nwr+=2;
        	      uN[j]=0;
        	    }
        	  }
        	  if (u) t0 = p->t;
        	  ok = 0;
        	}
        	break;
      }

      if (ok) {
        nwr += wdeltat(p->t,t0,Rtick);
	k = 0;
#ifdef NO_OFF
	if ((p->data[0]&0xf0)==0x90) {
	  if (p->data[0]==Rstatus) {
	  	k=1;
	  	Dprintf(" *");
	  }
	  Rstatus=p->data[0];
	}
	else Rstatus=0;
#endif
        while (k < p->len) {
          Dprintf(" %02x",p->data[k]);
          fputc(p->data[k],fd);
          k++;
          nwr++;
        }
        Dprintf("\n");
        t0 = p->t;
      }

      p = p->next;
    }
    for (j=0; j<128; j++) {
	if (uN[j]) {
	  Dprintf("CLOSE %x\n",j);
	  fputc(0,fd);
#ifdef NO_OFF
	  fputc(0x90|Cchan[i],fd);
#else
	  fputc(0x80|Cchan[i],fd);
#endif
	  fputc(j,fd); fputc(0,fd); nwr+=4;
	  uN[j]=0;
	}
    }
    fputc(0,fd);
    fputc(0xff,fd); fputc(0x2f,fd); fputc(0,fd); nwr+=4;
    fseek(fd,fptr,SEEK_SET);
    mwnum(nwr,4);
    fseek(fd,0,SEEK_END);
    cksum *= Rtick;
    Dprintf("CKSUM=%d\n",cksum);
    trkc++;
  }
  Dprintf("MThd: %d tracks written\n",trkc);
  return 0;
}


/* -------------------------------------------------------------------- */


int process(void)
{
  struct _htim cn[32];
  int vkeys[12];
  int i,j,k,nt,nch,mxc;
  float mnv;
  char *p;

  memset(Histo,0,sizeof(Histo));
  memset(Htone,0,sizeof(Htone));
  memset(cn,0,sizeof(cn));

  while (fgets(Line,250,fs)) {
    if (*Line=='#') {
      p=Line+3;
      if (!strncmp("Ins",p,3)) {
        p+=4; if (*p==' ') p++;
        if (sscanf(p,"%d",&i)!=1) continue;
        p=Line+32;
        if (sscanf(p,"%d %d",&j,&k)!=2) continue;
        Cmpu[i]=j;
        Cchan[i]=k;
      }
      continue;
    }
    if (!strncmp(Line,"Track ",6)) {
      Dprintf("\n%s",Line);
      sscanf(Line+6,"%d",&nt);
      if (nt > 0) {
        p = Line+8;
        if (*p=='(') {
          sscanf(p+1,"%d",&nch); p+=4;
        }
        else
          nch = nt-1;
        while (*p && isspace(*p)) p++;
        if (*p) {
          char *q;
          TrkNam[nch]=q=calloc(1,256);
          while (*p && !isspace(*p)) *q++=*p++;
          sprintf(q,"(%d)",nch+1);
        }
	scoreorder[nt-1]=nch;
	pass1(nch+1);
      }
      else {
        getzerotrk(fs);
      }
    }
  }
  Dprintf("\n\n");

  qsort(Histo,100,sizeof(struct _htim),hfc);
#ifdef DEBUG
  for (i=0; i<100; i++) {
    if (Histo[i].c) {
      Dprintf("%8.3f %3d\n",Histo[i].t*10.0,Histo[i].c);
    }
  }
  Dprintf("\n\n");
#endif

  for (i=0; i<32; i++) {
    float *pf = &Hdiv[i*32];
    float f;
    for (j=0; j<32; j++) {
      f = 32*Histo[i].t/Histo[j].t;
      if ((i!=j) && (((f-floor(f))<=0.1) || ((f-floor(f))>=0.9)))
        { pf[j]=f; cn[j].c += 1; cn[j].t += f; }
      else
        pf[j]=0;
    }
  }
  for (i=0; i<32; i++) {
    if (cn[i].c) cn[i].t /= cn[i].c;
  }

  mxc = 0; j = -1;
  mnv = 999999.9;
  for (i=0; i<32; i++) {
    if ((cn[i].c>=mxc) && (cn[i].t<mnv)) {
      j=i; mxc=cn[i].c; mnv=cn[i].t;
    }
  }
  Dprintf("j=%d c=%d v=%g\n",j,mxc,mnv*10.0);

  mnv = 999999.9; k = -1;
  for (i=0; i<32; i++) {
    float f = Hdiv[i*32+j];
    Dprintf("%3d %g\n",i,f);
    if ((f>0)&&(f<mnv)) { k=i; mnv=f; }
  }
  if (k<0) return 1;
  TimeUnit = (Histo[k].t*10.0);		/* ms */
  Dprintf("k=%d mcm=%d\n",k,TimeUnit);
  Dprintf("\n\n");

  for (i=0; i<12; i++) {
    k = Tkeys[i];
    vkeys[i] = 0;
    for (j=0; j<12; j++) {
      if (k&1) {
        vkeys[i] += Htone[j];
      }
      k >>= 1;
    }
    Dprintf("%2d %4d c=%d\n",i,Htone[i],vkeys[i]);
  }
  j=k=0;
  for (i=0; i<12; i++) {
    if (vkeys[i]>k) { j=i; k=vkeys[i]; }
  }
  Tonality = j-5;
  Dprintf("Tonality = %d\n",Tonality);
  Dprintf("\n\n");

  return 0;
}


/* -------------------------------------------------------------------- */

int main(int argc, char *argv[])
{
  char buf[256];
#ifdef NO_SYSEX
  char xbuf[256];
#endif
  char *p;
  int i,j;

  if (argc<2) return 1;
  fs = fopen(argv[1],"r"); if (fs==NULL) return 1;
  strcpy(buf,argv[1]);
  if ((p=strrchr(buf,'.'))!=NULL) *p=0;
#ifdef NO_SYSEX
  strcpy(xbuf,buf);
  strcat(xbuf,".syx");
#endif
  strcat(buf,".mid");

  for (i=0; i<64; i++) {
    Tracks[i] = Ttail[i] = (EVENT *)calloc(1,sizeof(EVENT));
    NEvTrk[i]=0;
    TrkNam[i]=NULL;
    Cmpu[i] = Cchan[i] = 0xff;
    scoreorder[i]=255;
  }
  ZeroTrk = ZeroTtail = (EVENT *)calloc(1,sizeof(EVENT));

#ifdef NO_SYSEX
  fx=fopen(xbuf,"w");
#endif
  process();

  fclose(fs);  
#ifdef NO_SYSEX
  fclose(fx);
#endif
  fd=fopen(buf,"w");

  genmidi();
  
  fclose(fd);

  for (i=0; i<64; i++) {
    EVENT *p = Tracks[i], *q;
    Dprintf("\n%d events in track %d MPU %d\n",NEvTrk[i],i,Cmpu[i]);
    while (p) {
      q=p; p=p->next;
#ifdef DEBUG
      if (q!=Tracks[i]) {
      	Dprintf("TK%02d t=%08d l=%2d\t",i,q->t,q->len);
      	for (j=0; j<q->len; j++) Dprintf(" %02x",(q->data[j])&0xff);
      	Dprintf("\n");
      }
#endif
      free(q);
    }
    if (TrkNam[i]) free(TrkNam[i]);
  }
  return 0;
}

