
/* ==================================================================== */
/*			RAW access to MPU-401				*/
/* ==================================================================== */

#include <asm/io.h>

#include "mpu401.h"


static __inline__ int putcmd(int n)
{
#ifdef MPU1
  int tout=0x7fff;

	while ((inb(MPU1+1)&0x40)&&(tout--));
	if (tout) outb(n,MPU1+1);
	  else return(-1);
	tout=0x7fff;
	while ((inb(MPU1+1)&0x80)&&(tout--));
	if (tout && (inb(MPU1)==0xfe)) return(0);
	  else return(-1);
#else
  return 0;
#endif
}

static __inline__ int putcmd2(int n)
{
#ifdef MPU2
  int tout=0x7fff;

	while ((inb(MPU2+1)&0x40)&&(tout--));
	if (tout) outb(n,MPU2+1);
	  else return(-1);
	tout=0x7fff;
	while ((inb(MPU2+1)&0x80)&&(tout--));
	if (tout && (inb(MPU2)==0xfe)) return(0);
	  else return(-1);
#else
  return 0;
#endif
}

static __inline__ int putcmd3(int n)
{
#ifdef MPU3
  int tout=0x7fff;

	while ((inb(MPU3+1)&0x40)&&(tout--));
	if (tout) outb(n,MPU3+1);
	  else return(-1);
	tout=0x7fff;
	while ((inb(MPU3+1)&0x80)&&(tout--));
	if (tout && (inb(MPU3)==0xfe)) return(0);
	  else return(-1);
#else
  return 0;
#endif
}

static __inline__ int putcmd4(int n)
{
#ifdef MPU4
  int tout=0x7fff;

	while ((inb(MPU4+1)&0x40)&&(tout--));
	if (tout) outb(n,MPU4+1);
	  else return(-1);
	tout=0x7fff;
	while ((inb(MPU4+1)&0x80)&&(tout--));
	if (tout && (inb(MPU4)==0xfe)) return(0);
	  else return(-1);
#else
  return 0;
#endif
}

/* ==================================================================== */

static int get401(int mpu)
{
 int i;

  while (1) {
#if VOICES>32
    switch (mpu) {
      case 3: i = getdat4(); break;
      case 2: i = getdat3(); break;
      case 1: i = getdat2(); break;
      default: i = getdata();
    }
#else
    i = (mpu? getdat2(): getdata());
#endif
    if (i != -1) return(i);
   }
}


static void putdata401(int mpu,int n)
{
#if VOICES>32
  switch (mpu) {
    case 3: putdat4(n); break;
    case 2: putdat3(n); break;
    case 1: putdat2(n); break;
    default: putdata(n);
  }
#else
  if (mpu) putdat2(n); else putdata(n);
#endif
}


static void sxcmd(int mpu,int c)
{
 static char Cb[] = { 0xf0,0x18,0x04,0x00 };
 int i;

  for (i=0; i<4; i++) putdata401(mpu,Cb[i]);
  putdata401(mpu,c);
}


static void eox(int mpu)
{
  putdata401(mpu,0xf7);
}


static int sendcmd401(int mpu,int n)
{
 int ans;

#if VOICES>32
  switch (mpu) {
    case 3: ans = putcmd4(n); break;
    case 2: ans = putcmd3(n); break;
    case 1: ans = putcmd2(n); break;
    default: ans = putcmd(n);
  }
#else
  ans = (mpu? putcmd2(n):putcmd(n));
#endif
  if (ans == ACK) return(ans);
   else if (ans != -1) {
     /* ungetnext401(ans);      put pending data on stack */
     while (1) {              /* check for more incoming data */
#if VOICES>32
       switch (mpu) {
	 case 3: ans = putcmd4(n); break;
	 case 2: ans = putcmd3(n); break;
	 case 1: ans = putcmd2(n); break;
	 default: ans = putcmd(n);
       }
#else
       ans = (mpu? getdat2():getdata());
#endif
       if (ans == ACK || ans == -1) return(ans);
	/* else ungetnext401(ans); */
      }
    }
  return(ans);
}


int repeat_cmd401(int mpu,int n)
{
 int i, j, m;

  for (i = 0; i < 10; i++){
    m = sendcmd401(mpu,n);
    if (m != -1) return(m);
     else {
       for (j = 0; j < 10; j++) {
#if VOICES>32
	 switch (mpu) {
	   case 3: getdat4(); break;
	   case 2: getdat3(); break;
	   case 1: getdat2(); break;
	   default: getdata();
	 }
#else
	 if (mpu) getdat2(); else getdata();
#endif
	}
      }
   }
  return(-1);
}


void all_notes_off(int mpu)
{
 int i;

#if !defined(DTC2) && !defined(DTC3)
  sendcmd401(mpu,RESET);
  usleep(10000);

  repeat_cmd401(mpu,UART);
#endif
  for(i = 0; i < 16; i++){
    putdata401(mpu,MODE_MESSAGE | i);
    putdata401(mpu,0x7f);
    putdata401(mpu,0);
    putdata401(mpu,MODE_MESSAGE | i);
    putdata401(mpu,0x79);
    putdata401(mpu,0);
    putdata401(mpu,MODE_MESSAGE | i);
    putdata401(mpu,ALL_NOTES_OFF);
    putdata401(mpu,0);
    usleep(10000);
   }
#if !defined(DTC2) && !defined(DTC3)
  sendcmd401(mpu,RESET);
#endif
}

/* ==================================================================== */

