
/* ================================================================== */
/*			PLAYER utilities for Tx                       */
/*			  by Harawi 9/3/93                            */
/* ================================================================== */

#include "mplay.h"

/* ================================================================== */

unsigned Bittab[32] =
	{ 0x00000001,0x00000002,0x00000004,0x00000008,
	  0x00000010,0x00000020,0x00000040,0x00000080,
	  0x00000100,0x00000200,0x00000400,0x00000800,
	  0x00001000,0x00002000,0x00004000,0x00008000,
	  0x00010000,0x00020000,0x00040000,0x00080000,
	  0x00100000,0x00200000,0x00400000,0x00800000,
	  0x01000000,0x02000000,0x04000000,0x08000000,
	  0x10000000,0x20000000,0x40000000,0x80000000 };

/* ================================================================== */
/*
 * 1/f noise according to Voss' algorithm.
 * Sequence length (Fseq) may be any number, but is truncated to the
 * nearest power of 2.
 * The algorithm:
 * Select log2(sequence) separate white noise generators, i.e.,
 * one for each bit of a binary number.
 * Initialize a counting register.
 * In a loop, add 1 to the counting register, determine which
 * bits have changed from its previous value.  For each changed
 * bit, invoke the appropriate generator to produce a new number.
 * Sum the normalized output of all generators as the current number.
 * The procedure f1init() is used to reset the sequence length.
 * On the next call to onefrand(), the new value of Fseq will be used.
 */

static int   f1ini;
FLOAT Fseq = 32.0 * 1024.0;


FLOAT onefrand (FLOAT lb, FLOAT ub)
{
 int i, dif;
 static int gens;
 static FLOAT *sums;
 FLOAT  sum;
 static FLOAT div;
 static int cnt = -1, ocnt;

  if (f1ini == 0) {
    f1ini++;
    if (Fseq <= 1.0) Fseq = 2.0;
    gens = log(Fseq) / log(2.0);
    if (sums != NULL) free(sums);
    sums = (FLOAT *) malloc(sizeof(FLOAT) * gens);
    for (i = 0; i < gens; i++) sums[i] = 0;
    div = pow(2.0, (FLOAT) ((sizeof(int) * 8) - 1)) - 1.0;
  }
  ocnt = cnt++;			/* increment count */
  dif = cnt ^ ocnt;		/* set sums to change */

  for (i = 0; i < gens; i++) {	/* form new sums */
    if (dif & (1 << i))
      sums[i] = rand();
  }

  for (sum = i = 0; i < gens; i++)	/* sum them all */
    sum += sums[i];

  sum /= gens * div;
  sum = (sum * (ub - lb)) + lb;	/* norm to spec */
  return (sum);
}


void f1init (FLOAT sequence)
{
  f1ini = 0;
  Fseq = sequence;
}


/* ================================================================== */


static float Mcr=0.5,Mci=0.5;


Pair *rmandel(int mn,int mx)
{
 static float ar=0.0,ai=0.0;
 float t,r2,i2;
 static Pair p;

  r2=ar*ar; i2=ai*ai;
  t=r2-i2+Mcr;
  ai=2*ar*ai+Mci;
  ar=t;
  if ((r2+i2)>2) { ar=0.0; ai=0.0; }
  p.x = ar*(mx-mn); if (p.x<0) p.x=mn-p.x; else p.x+=mn;
  p.y = ai*(mx-mn); if (p.y<0) p.y=mn-p.y; else p.y+=mn;
  return(&p);
}


void mbinit(float r,float i)
{
  Mcr=r; Mci=i;
}


/* ================================================================== */

#define MULTIPLIER      0x015a4e35L
#define INCREMENT       1

unsigned Rseed = 1;

void Srand(unsigned seed)
{
  Rseed = seed;
}

unsigned Rand(void)
{
  Rseed = MULTIPLIER * Rseed + INCREMENT;
  return(Rseed>>4);
}

/* ================================================================== */

